#include <iostream>
#include <iomanip>
#include <chrono>

#include "QWindow.hpp"

#include "Config.hpp"
using namespace QPong;

bool propagating = true;
int propRate = 0;
int renderRate = 0;

bool qPressed = false;
bool aPressed = false;
bool upPressed = false;
bool downPressed = false;

/// @brief This functino has to run in its own thread and propagates the particle as fast as possible.
///
void QWindow::simulation()
{
    do
    {
        auto start = std::chrono::system_clock::now();

        m_particle->propagate();
        propRate = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now() - start).count();
            
        if (qPressed)
            m_players[Player::Id::One]->move(Player::Direction::Up);

        if (aPressed)
            m_players[Player::Id::One]->move(Player::Direction::Down);

        if (upPressed)
            m_players[Player::Id::Two]->move(Player::Direction::Up);

        if (downPressed)
            m_players[Player::Id::Two]->move(Player::Direction::Down);

        // Limit the speed of the simulation.
        auto dt = std::chrono::system_clock::now() - start;
        auto dtMillis = std::chrono::duration_cast<std::chrono::milliseconds>(dt).count();
        constexpr int propTime = 6;
        if (dtMillis < propTime)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(propTime - dtMillis));
        }
    } 
    while (propagating);
}

/// @brief This function has to run in its own thread an will redraw the canvas every ~25ms.
///
void render(QWindow *w)
{
    do
    {
        using namespace std::chrono_literals;
        auto start = std::chrono::system_clock::now();
        std::this_thread::sleep_for(15ms);
        w->updateView();
        renderRate = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now() - start).count();
    } while (true);
}

QWindow::QWindow() 
    : m_button{"Initialise Particle"}
    , m_label("Hi, you are playing QPong and this is a label!")
    , m_scorePlayerOne("0")
    , m_scorePlayerTwo("0")
{
    set_border_width(5);

    constexpr double playerOneStartX = -0.7;
    constexpr double playerTwoStartX = 0.7;
    m_players.insert(std::make_pair<Player::Id, std::shared_ptr<Player>>(Player::Id::One, std::make_shared<Player>(Player::Id::One, Player::Position{-0.7, 0})));
    m_players.insert(std::make_pair<Player::Id, std::shared_ptr<Player>>(Player::Id::Two, std::make_shared<Player>(Player::Id::Two, Player::Position{ 0.7, 0.4})));

    m_momentum = new ParticleImage;
    m_position = new ParticleImage;
    m_particle = new Particle(m_momentum, m_position, m_players);

    add(m_box);
    m_box.pack_start(m_vBox);
    m_vBox.pack_start(m_button, Gtk::PACK_SHRINK);
    m_vBox.pack_start(m_label, Gtk::PACK_SHRINK);
    
    m_fpsBox.pack_start(m_renderRate, Gtk::PACK_EXPAND_WIDGET);
    m_fpsBox.pack_start(m_proppergatingRate, Gtk::PACK_EXPAND_WIDGET);
    m_vBox.pack_start(m_fpsBox, Gtk::PACK_SHRINK);

    m_scoreBox.pack_start(m_scorePlayerOne, Gtk::PACK_EXPAND_WIDGET);
    m_scoreBox.pack_start(m_scorePlayerTwo, Gtk::PACK_EXPAND_WIDGET);
    m_vBox.pack_start(m_scoreBox, Gtk::PACK_SHRINK);

    m_hBox.pack_start(m_positionArea, Gtk::PACK_EXPAND_WIDGET);
    m_hBox.pack_start(m_momentumArea, Gtk::PACK_EXPAND_WIDGET);
    m_vBox.pack_start(m_hBox);

    m_label.set_margin_top(5);  
    m_label.set_margin_bottom(5);
    m_proppergatingRate.set_text("0");
    m_renderRate.set_text("0");
    m_renderRate.set_margin_top(5);  
    m_renderRate.set_margin_bottom(5);
    m_proppergatingRate.set_margin_top(5);  
    m_proppergatingRate.set_margin_bottom(5);

    m_button.signal_clicked().connect(sigc::mem_fun(*this, &QWindow::onButtonPress));
    
    this->signal_key_press_event().connect( sigc::mem_fun( *this, &QWindow::onKeyPress));
    this->signal_key_release_event().connect(sigc::mem_fun(*this, &QWindow::onKeyPress));

    m_guiDispatcher.connect(sigc::mem_fun(*this, &QWindow::updateGui));
    m_areaDispatcher.connect(sigc::mem_fun(*this, &QWindow::updateImage));

    while (m_particle->notReady())
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }

    m_propagatingThread = std::thread(&QWindow::simulation, this);
    m_propagatingThread.detach();

    std::thread rendering(render, this);
    rendering.detach();

    m_box.show_all();
}

unsigned guiUpdates = 0;
void QWindow::updateGui()
{
    guiUpdates++;
    if (guiUpdates % 10 == 0)
    {
        std::stringstream ss;
        ss << "image update rate: ";
        ss << 1000 / renderRate;
        m_renderRate.set_text(ss.str());
        ss = std::stringstream();
        ss << "simulation rate: ";
        ss << 1000 / propRate;
        m_proppergatingRate.set_text(ss.str());

        ss = std::stringstream();
        ss << "Score Player 1: ";
        auto score = m_players[Player::Id::One]->getScore();
        ss << std::setprecision(2) << std::setw(7) << score << "%";
        if (score > 50.0)
        {
            ss << std::endl << "GEWONNEN!";
            ss << std::endl << "Zum Neustarten Partikel neu initialisieren";
        }
        m_scorePlayerOne.set_text(ss.str());

        ss = std::stringstream();
        ss << "Score Player 2: ";
        score = m_players[Player::Id::Two]->getScore();
        ss << std::setprecision(2)  << std::setw(7) << score << "%";
        if (score > 50.0)
        {
            ss << std::endl << "GEWONNEN!";
            ss << std::endl << "Zum Neustarten Partikel neu initialisieren";
        }
        m_scorePlayerTwo.set_text(ss.str());
    }
}


bool QWindow::onKeyPress(GdkEventKey* event)
{
    constexpr int q = 113; 
    constexpr int a = 97;
    constexpr int ue = 252;
    constexpr int oe = 246;

    if (event->type == GdkEventType::GDK_KEY_PRESS)
    {
        if (event->keyval == q)
            qPressed = true;
        if (event->keyval == a)
            aPressed = true;
        if (event->keyval == ue)
            upPressed = true;
        if (event->keyval == oe)
            downPressed = true;
    }
    else if (event->type == GdkEventType::GDK_KEY_RELEASE)
    {
        if (event->keyval == q)
            qPressed = false;
        if (event->keyval == a)
            aPressed = false;
        if (event->keyval == ue)
            upPressed = false;
        if (event->keyval == oe)
            downPressed = false;
    }
    return true;
}

void QWindow::updateView()
{
    m_areaDispatcher.emit();
    m_guiDispatcher.emit();
}

void QWindow::updateImage()
{
    m_positionArea.setPixbuf(m_position->getPixbuf());
    m_momentumArea.setPixbuf(m_momentum->getPixbuf());
    m_positionArea.redraw();
    m_momentumArea.redraw();
}   

void QWindow::onButtonPress()
{
    propagating = false;

    Config::loadFrom("qpong.config");
    
    std::this_thread::sleep_for((std::chrono::seconds)2);

    delete(m_particle);
    delete(m_position);
    delete(m_momentum);
    m_position = new ParticleImage;
    m_momentum = new ParticleImage;
    m_particle = new Particle(m_momentum, m_position, m_players);

    while (m_particle->notReady())
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }
    
    for (auto player : m_players)
    {
        player.second->reset();
    }

    m_propagatingThread = std::thread(&QWindow::simulation, this);

    propagating = true;
    m_propagatingThread.detach();
}