///
/// @file       Particle.hpp
/// @author     Armin Co
/// @brief      Particle class header.
///

#ifndef QPONG_PARTICLE_HPP
#define QPONG_PARTICLE_HPP

#include <complex>
#include <memory>
#include <map>

#include <math.h>
#include <fftw3.h>

#include "ImageBuffer.hpp"
#include "Player.hpp"

namespace QPong {

class Particle
{
public:

    /// @brief Contructor
    ///
    Particle(ImageBuffer *momentum, ImageBuffer *position, std::map<Player::Id, std::shared_ptr<Player>> players);
    
    
    /// @brief Destructor
    ///
    ~Particle();


    /// @brief Execute one step of the simulation.
    ///
    void propagate();


    /// @brief Init the array with initial momentum.
    ///        Representation is in position representation.
    /// 
    void initMomentum();


    /// @brief Check if everything is initialised.
    ///
    bool notReady();


private:
    /// @brief Update the array at the ArrayCanvas
    ///
    void updateMomentumImage();


    /// @brief Manipulate particle in position representation
    ///        For bats and borders potentials are used.
    ///        Removing the particle behind the bat with a cos2
    ///
    void manipulateParticleInPosition(int from, int to);


    /// @brief Move the particle one step forward in momentum representation
    ///
    void moveStep(int indexFrom, int indexTo);


    /// @brief Remove parts of the particle that are behind the bat.
    ///
    double removeParticle(int index, double cx);


    std::complex<double> *m_psi;
    fftw_plan m_planForward;
    fftw_plan m_planBackward;
    bool m_ready;

    std::map<Player::Id, std::shared_ptr<Player>> m_players;

    ImageBuffer *m_bufPositionRepresentation;
    ImageBuffer *m_bufMomentumRepresentation;




    Particle() = delete;
};

} // namespace QPong

#endif